from collections import deque
from types import MethodType


class Data():
    pass


class Loader():

    def __init__(self, session):
        self.session = session
        self.loading_queue = deque()
        self.queued_pieces = {}
        self.proxies = {}

    def load(self, sets):

        data = Data()

        for scenario_class in sets:
            setattr(data, scenario_class.__name__, Data())
            for piece in self.get_pieces(scenario_class()):
                self.queue(piece)

        while True:
            try:
                piece = self.loading_queue.popleft()
            except IndexError:
                break

            self.load_piece(piece, data)

        self.session.expunge_all()
        self.queued_pieces = {}
        self.proxies = {}

        return data

    def full_name(self, piece):
        return ".".join([
            piece.im_class.__name__, piece.__name__
        ])

    def load_piece(self, piece, data):
        model_data = getattr(piece.im_class(), piece.__name__)()
        model_class = piece.proxy_model_class
        for reference in self.get_references(model_data):
            referenced_model = self.proxies[
                self.full_name(reference)
            ]
            for foreign_key in model_class.__table__.foreign_keys:
                if foreign_key.references(referenced_model.__table__):
                    model_data[
                        foreign_key.constraint.columns[0].name
                    ] = getattr(
                        referenced_model, foreign_key.column.name
                    )

        model = model_class(
            **{
                column_name: value
                for (column_name, value) in model_data.iteritems()
                if type(value) != MethodType
            }
        )
        self.session.add(model)
        self.session.flush()
        self.session.refresh(model)

        self.proxies[
            self.full_name(piece)
        ] = model

        for column in model.__table__.columns:
            model_data[column.name] = getattr(model, column.name)

        piece_data = Data()

        setattr(
            getattr(data, piece.im_class.__name__),
            piece.__name__,
            piece_data
        )
        for column_name, value in model_data.iteritems():
            if type(value) == MethodType:
                value = getattr(
                    getattr(data, value.im_class.__name__),
                    value.__name__
                )

            setattr(piece_data, column_name, value)

    def get_pieces(self, scenario_class):
        pieces = []
        for slot in dir(scenario_class):
            if type(getattr(scenario_class, slot)) == MethodType:
                pieces.append(getattr(scenario_class, slot))

        return pieces

    def get_references(self, model_data):
        references = []
        for value in model_data.values():
            if type(value) == MethodType:
                references.append(value)

        return references

    def is_queued(self, piece):
        return self.queued_pieces.get(
            self.full_name(piece),
            False
        )

    def queue(self, piece):
        if piece.im_self is None:
            model_data = piece(piece.im_class())
        else:
            model_data = piece()
        for reference in self.get_references(model_data):
            if not self.is_queued(reference):
                self.queue(reference)

        if not self.is_queued(piece):
            self.queued_pieces[
                self.full_name(piece)
            ] = True
            self.loading_queue.append(piece)

__all__ = [Loader]
